import enum
import threading
import time
from ctypes import *
from ctypes.wintypes import *
from functools import partial
from typing import Callable, Optional, Tuple

from .winevent import HWINEVENTHOOK, WINEVENTHOOKPROC, WinEvent

user32 = WinDLL("user32", use_last_error=True)

HC_ACTION = 0
# incoming message
WM_QUIT = 0x0012


# types for the hook including input parameter and return result

ULONG_PTR = WPARAM
LRESULT = LPARAM
# LPMSG = POINTER(MSG)

# Hook
HOOKPROC = WINFUNCTYPE(LRESULT, c_int, WPARAM, LPARAM)


# setup api


def _errcheck_bool(result, func, args):
    if not result:
        raise WinError(get_last_error())
    return args


user32.SetWindowsHookExW.errcheck = _errcheck_bool
user32.SetWindowsHookExW.restype = HHOOK
user32.SetWindowsHookExW.argtypes = (
    c_int,  # _In_ idHook
    HOOKPROC,  # _In_ lpfn
    HINSTANCE,  # _In_ hMod
    DWORD,
)  # _In_ dwThreadId

user32.CallNextHookEx.restype = LRESULT
user32.CallNextHookEx.argtypes = (
    HHOOK,  # _In_opt_ hhk
    c_int,  # _In_     nCode
    WPARAM,  # _In_     wParam
    LPARAM,
)  # _In_     lParam

user32.GetMessageW.argtypes = (
    LPMSG,  # _Out_    lpMsg
    HWND,  # _In_opt_ hWnd
    UINT,  # _In_     wMsgFilterMin
    UINT,
)  # _In_     wMsgFilterMax

user32.TranslateMessage.argtypes = (LPMSG,)
user32.DispatchMessageW.argtypes = (LPMSG,)

user32.SetWinEventHook.restype = HWINEVENTHOOK

# keyboard hook definition


class KBDLLHOOKMSGID(enum.IntEnum):
    """Keyboard event msgid"""

    WM_KEYDOWN = 0x0100
    WM_KEYUP = 0x0101
    WM_SYSKEYDOWN = 0x0104
    WM_SYSKEYUP = 0x0105


class KBDLLHOOKDATA(Structure):
    """Keyboard event data"""

    vkCode: int
    scanCode: int
    flags: int
    time: int
    dwExtraInfo: int

    _fields_ = (
        ("vkCode", DWORD),
        ("scanCode", DWORD),
        ("flags", DWORD),
        ("time", DWORD),
        ("dwExtraInfo", ULONG_PTR),
    )


# mouse hook definition


class MSLLHOOKMSGID(enum.IntEnum):
    """Mouse event msgid"""

    WM_MOUSEMOVE = 0x0200
    WM_LBUTTONDOWN = 0x0201
    WM_LBUTTONUP = 0x0202
    WM_RBUTTONDOWN = 0x0204
    WM_RBUTTONUP = 0x0205
    WM_MBUTTONDOWN = 0x0207
    WM_MBUTTONUP = 0x0208
    WM_MOUSEWHEEL = 0x020A
    WM_MOUSEHWHEEL = 0x020E


class MSLLHOOKDATA(Structure):
    """Mouse event data"""

    pt: POINT
    mouseData: int
    flags: int
    time: int
    dwExtraInfo: int

    _fields_ = (
        ("pt", POINT),
        ("mouseData", DWORD),
        ("flags", DWORD),
        ("time", DWORD),
        ("dwExtraInfo", ULONG_PTR),
    )


# wrap them all inside Hook class


def keyboard(msgid: KBDLLHOOKMSGID, msg: KBDLLHOOKDATA) -> bool:
    print(
        "{:15s} {:15s}: vkCode {:3x} scanCode {:3x} flags: {:3d}, time: {} extra: {}".format(
            Vk(msg.vkCode).name,
            msgid.name,
            msg.vkCode,
            msg.scanCode,
            msg.flags,
            msg.time,
            msg.dwExtraInfo,
        )
    )


class Hook(threading.Thread):
    """Hook low level keyboard/mouse event

    Usage:

    .. code-block:: python

        def swallow_keyboard_a_key_event(msgid: KBDLLHOOKMSGID, data: KBDLLHOOKDATA) -> bool:
            return data.vkCode == VirtualKey.VK_A:

        def swallow_mouse_middle_btn_event(msgid: MSLLHOOKMSGID, data: MSLLHOOKDATA) -> bool:
            return msgid == MSLLHOOKMSGID.WM_MBUTTONUP or msgid == MSLLHOOKMSGID.WM_MBUTTONDOWN:

        def winevent(
            event: WinEvent,
            hwnd: HWND,
            id_obj: LONG,
            id_chd: LONG,
            id_evt_thread: DWORD,
            time: DWORD,
        ):
            pass


        hook = Hook()
        kbd_hook_id = hook.install_keyboard_hook(swallow_keyboard_a_key_event)
        mouse_hook_id = hook.install_mouse_hook(swallow_mouse_middle_btn_event)
        winevent_hook_id = hook.install_winevent_hook()
        hook.start()

    """

    def __init__(self):
        self._installed_hooks = {}
        self._queue = []
        super().__init__()

    def _install_hook(
        self, hook_id, wparam_type, lparam_type, handler
    ) -> Tuple[HANDLE, Callable]:
        # for the hooks to work, note that only low level keyboard/mouse work this way
        # while others require DLL injection
        @HOOKPROC
        def proc(nCode, wParam, lParam):
            if nCode == HC_ACTION:
                wparam = wparam_type(wParam)
                lparam = cast(lParam, POINTER(lparam_type))[0]
                if handler(wparam, lparam):
                    return 1
            return user32.CallNextHookEx(None, nCode, wParam, lParam)

        hook = user32.SetWindowsHookExW(hook_id, proc, None, 0)
        return hook, proc

    def _install_winevent_hook(
        self,
        event_min: WinEvent,
        event_max: WinEvent,
        cb: Callable[[WinEvent, HWND, LONG, LONG, DWORD, DWORD], None],
    ) -> Tuple[HANDLE, Callable]:
        @WINEVENTHOOKPROC
        def proc(hhook, event, hwnd, id_obj, id_chd, id_evt_thread, dwms_evt_time):
            cb(WinEvent(event), hwnd, id_obj, id_chd, id_evt_thread, dwms_evt_time)

        hook = user32.SetWinEventHook(
            DWORD(event_min.value),
            DWORD(event_max.value),
            HMODULE(None),
            proc,
            0,
            0,
            0,  # WINEVENT_OUTOFCONTEXT
        )
        return hook, proc

    def install_keyboard_hook(
        self, callback: Callable[[KBDLLHOOKMSGID, KBDLLHOOKDATA], bool]
    ):
        """Install keyboard hook

        :param callback: function to be called when key press/release, return ``True`` to stop
            propagation
        """
        self._queue.append(
            partial(self._install_hook, 13, KBDLLHOOKMSGID, KBDLLHOOKDATA, callback)
        )

    def install_mouse_hook(
        self, callback: Callable[[MSLLHOOKMSGID, MSLLHOOKDATA], bool]
    ):
        """Install mouse hook

        :param callback: function to be called when mouse moved, bth pressed/release and scroll,
 return ``True`` to stop
            propagation
        """
        self._queue.append(
            partial(self._install_hook, 14, MSLLHOOKMSGID, MSLLHOOKDATA, callback)
        )

    def install_winevent_hook(
        self,
        callback: Callable[[WinEvent, HWND, LONG, LONG, DWORD, DWORD], None],
        event_min: WinEvent,
        event_max: Optional[WinEvent] = None,
    ):
        """Install winevent hook

        :param callback: function to be called when Windows event occurred
        """
        self._queue.append(
            partial(
                self._install_winevent_hook, event_min, event_max or event_min, callback
            )
        )

    def run(self):
        """Start hooking, all installed hooks would not actually work until
        this method invoked. You may install new hooks thereafter"""
        # IMPORTANT: the hook must be installed in the thread!
        msg = MSG()
        while True:
            # install pending hooks
            while self._queue:
                install_hook = self._queue.pop()
                hook, proc = install_hook()
                self._installed_hooks[hook] = proc

            bRet = user32.GetMessageW(byref(msg), None, 0, 0)
            if not bRet:
                break
            if bRet == -1:
                raise WinError(get_last_error())
            user32.TranslateMessage(byref(msg))
            user32.DispatchMessageW(byref(msg))

    def stop(self):
        """Stop all hooks"""
        user32.PostThreadMessageW(self.ident, WM_QUIT, 0, 0)


if __name__ == "__main__":
    from datetime import datetime

    from . import window
    from .vk import Vk

    def keyboard(msgid: KBDLLHOOKMSGID, msg: KBDLLHOOKDATA) -> bool:
        print(
            "{:15s} {:15s}: vkCode {:3x} scanCode {:3x} flags: {:3d}, time: {} extra: {}".format(
                Vk(msg.vkCode).name,
                msgid.name,
                msg.vkCode,
                msg.scanCode,
                msg.flags,
                msg.time,
                msg.dwExtraInfo,
            )
        )

    def mouse(msgid: MSLLHOOKMSGID, msg: MSLLHOOKDATA) -> bool:
        msg = (
            (msg.pt.x, msg.pt.y),
            msg.mouseData,
            msg.flags,
            msg.time,
            msg.dwExtraInfo,
        )
        print("{:15s}: {}".format(msgid.name, msg))

    def winevent(
        event: WinEvent,
        hwnd: HWND,
        id_obj: LONG,
        id_chd: LONG,
        id_evt_thread: DWORD,
        time: DWORD,
    ):
        print("==================================")
        # print(
        #     hwnd is None,
        #     id_obj is None,
        #     id_chd is None,
        #     window.Window(hwnd).title is None,
        # )
        print(
            "[{now}] {event:30s} {hwnd:8d} ido: {id_obj:6d} idc: {id_chd:6d} {title}".format(
                now=datetime.now().strftime("%M:%S.%f"),
                event=event.name,
                hwnd=hwnd or 0,
                id_obj=id_obj,
                id_chd=id_chd,
                title=window.Window(hwnd).title,
            )
        )
        window.inspect_window(hwnd)
        print("==================================")

    hook = Hook()
    # hook.install_winevent_hook(winevent, WinEvent.EVENT_SYSTEM_MOVESIZEEND)
    # hook.install_winevent_hook(winevent, WinEvent.EVENT_SYSTEM_MINIMIZESTART, WinEvent.EVENT_SYSTEM_MINIMIZEEND)
    # hook.install_winevent_hook(
    #     winevent, WinEvent.EVENT_OBJECT_SHOW, WinEvent.EVENT_OBJECT_HIDE
    # )
    # hook.install_winevent_hook(
    #     winevent, WinEvent.EVENT_OBJECT_CREATE, WinEvent.EVENT_OBJECT_DESTROY
    # )
    hook.install_winevent_hook(winevent, WinEvent.EVENT_MIN, WinEvent.EVENT_MAX)
    # hook.install_keyboard_hook(keyboard)
    hook.start()

    while True:
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            hook.stop()
            break



import enum
import typing
from ctypes import *
from ctypes.wintypes import *

user32 = WinDLL("user32", use_last_error=True)

ULONG_PTR = LPARAM


class INPUTTYPE(enum.IntEnum):
    MOUSE = 0
    KEYBOARD = 1
    HARDWARE = 2


class KEYEVENTF(enum.IntFlag):
    EXTENDEDKEY = 0x0001
    KEYUP = 0x0002
    SCANCODE = 0x0008
    UNICODE = 0x0004


class KEYBDINPUT(Structure):
    _fields_ = (
        ("wVk", WORD),
        ("wScan", WORD),
        ("dwFlags", DWORD),
        ("time", DWORD),
        ("dwExtraInfo", ULONG_PTR),
    )


class MOUSEINPUT(Structure):
    _fields_ = (
        ("dx", LONG),
        ("dy", LONG),
        ("mouseData", DWORD),
        ("dwFlags", DWORD),
        ("time", DWORD),
        ("dwExtraInfo", ULONG_PTR),
    )


class HARDWAREINPUT(Structure):
    _fields_ = (
        ("uMsg", LONG),
        ("wParamL", WORD),
        ("wParamH", WORD),
    )


class INPUTDATA(Union):
    _fields_ = (
        ("mi", MOUSEINPUT),
        ("ki", KEYBDINPUT),
        ("hi", HARDWAREINPUT),
    )


class INPUT(Structure):
    _anonymous_ = ["u"]
    _fields_ = (
        ("type", DWORD),
        ("u", INPUTDATA),
    )


SYNTHESIZED_ID = 123123123


def send_input(*inputs: typing.List[INPUT]):
    global SYNTHESIZED_ID
    """Synthesizes keystrokes, mouse motions, and button clicks.

    Usage:
    ```
    send_input(
        INPUT(
            type=INPUTTYPE.KEYBOARD,
            ki=KEYBDINPUT(wVk=VirtualKey.A),
        ),
        INPUT(
            type=INPUTTYPE.KEYBOARD,
            ki=KEYBDINPUT(wVk=VirtualKey.A, dwFlags=KEYEVENTF.KEYUP),
        ),
    )
    ```

    :param List[INPUT] inputs: list of keyboard/mouse inputs which
        to be sent to system

    """
    for item in inputs:
        if item.ki:
            item.ki.dwExtraInfo = SYNTHESIZED_ID
        elif item.mi:
            item.mi.dwExtraInfo = SYNTHESIZED_ID
    length = len(inputs)
    array = INPUT * length
    if not user32.SendInput(length, array(*inputs), sizeof(INPUT)):
        raise WinError(get_last_error())


def is_synthesized(msg: typing.Union[KEYBDINPUT, MOUSEINPUT]) -> bool:
    """Check if keyboard/mouse event is sent by this module
    """
    global SYNTHESIZED_ID
    return msg.dwExtraInfo == SYNTHESIZED_ID


if __name__ == "__main__":
    from .vk import Vk

    send_input(
        INPUT(
            type=INPUTTYPE.KEYBOARD,
            ki=KEYBDINPUT(wVk=Vk.A),
        ),
        INPUT(
            type=INPUTTYPE.KEYBOARD,
            ki=KEYBDINPUT(wVk=Vk.A, dwFlags=KEYEVENTF.KEYUP),
        ),
    )



from enum import Enum
from subprocess import PIPE, Popen
from threading import Lock
from typing import List, Optional, TextIO


class ServiceStatus(Enum):
    STOPPED = "stopped"
    RUNNING = "running"


class ServiceEntry:
    """ServiceEntry represents a service, holds the meta data and running state

    :param str name: name of the service
    :param str desc: description
    :param List[str] args: the arguments used to launch the process.
    :param bool autostart: start the service automatically
    :param str log_path: log file path
    :param bool log_append_only: append log to the file or overwrite
    """

    name: str
    desc: str
    args: List[str]
    autostart: bool
    log_path: str
    log_append_only: bool
    _process: Popen = None
    _log_file: TextIO = None
    _lock: Lock

    def __init__(
        self,
        name,
        args,
        log_path=None,
        log_append_only=False,
        desc=None,
        autostart=True,
    ):
        self.name = name
        self.desc = desc
        self.args = args
        self.log_path = log_path
        self.autostart = autostart
        self.log_append_only = log_append_only
        self._lock = Lock()

    @property
    def status(self) -> ServiceStatus:
        """Retrieves the status of the service"""
        with self._lock:
            if self._process is None:
                return ServiceStatus.STOPPED
            if self._process.poll() is None:
                return ServiceStatus.RUNNING
        return ServiceStatus.STOPPED

    def start(self):
        """Start the service"""
        with self._lock:
            if self._process is not None:
                raise ValueError(f"Service {self.name} is already running")
            log_file = PIPE
            if self.log_path:
                log_file = open(self.log_path, "a+" if self.log_append_only else "w+")
                self._log_file = log_file
            self._process = Popen(
                self.args, stdout=log_file, stdin=log_file, shell=True
            )

    @property
    def status_text(self) -> str:
        """Retrieves the status of the service in plaintext format"""
        return "running" if self.status == ServiceStatus.RUNNING else "stopped"

    def stop(self):
        """Stop the service"""
        with self._lock:
            if self._process is None:
                return
            Popen("TASKKILL /F /PID {pid} /T".format(pid=self._process.pid))
            self._process = None
            if self._log_file:
                self._log_file.close()
            self._log_file = None

    def restart(self):
        """Restart the service"""
        self.stop()
        self.start()

    def toggle(self):
        """Start or stop the service"""
        if self.status == ServiceStatus.STOPPED:
            self.start()
        else:
            self.stop()


class ServiceManager:
    """Simple service manager to help you run programs as a service, especially
    helpful for console software like `syncthing`."""

    _services: List[ServiceEntry]
    _lock: Lock

    def __init__(self):
        self._services = []
        self._lock = Lock()

    def find_by_name(self, name: str) -> Optional[ServiceEntry]:
        """Find a service by name, return None if not found"""
        for service in self._services:
            if service.name == name:
                return service

    def register(self, service: ServiceEntry):
        """Register a new service"""
        with self._lock:
            if self.find_by_name(service.name):
                raise ValueError(f"Service {service.name} already exists")
            self._services.append(service)
            service.start()

    def start(self):
        """Starts all services that have autostart enabled"""
        with self._lock:
            for service in self._services:
                if service.autostart:
                    service.restart()

    def stop(self):
        """Stops all running services"""
        with self._lock:
            for service in self._services:
                service.stop()

    def get_all(self) -> List[ServiceEntry]:
        """Retrieves a copy of all registered services"""
        with self._lock:
            return self._services.copy()


svcmgr = ServiceManager()

register_service = svcmgr.register
get_services = svcmgr.get_all
start_services = svcmgr.start
stop_services = svcmgr.stop



import json
import os
from dataclasses import dataclass
from datetime import datetime
from typing import Callable, Dict, List, Union

from jigsawwm.state import get_state_manager
from jigsawwm.w32.process import is_exe_running


def open_chrome_fav_folder(root, fav_folder):
    """Opens the Chrome Favorites folder"""
    path = os.path.join(
        os.getenv("LOCALAPPDATA"),
        "Google",
        "Chrome",
        "User Data",
        "Default",
        "Bookmarks",
    )
    bookmarks = None
    with open(path, encoding="utf8") as f:
        bookmarks = json.load(f)

    # open folder recursively
    def open_folder(folder):
        for child in folder["children"]:
            if child.get("type") == "url":
                os.startfile(child["url"])
            else:
                open_folder(child)

    # locate the initial folder
    folder = bookmarks["roots"][root]
    for component in fav_folder.split("."):
        for child in folder["children"]:
            if child["name"] == component and child["type"] == "folder":
                folder = child
                break
    open_folder(folder)


def daily_once(name: str):
    """Returns True if the given name has not been called today"""
    today = datetime.today().date()
    state = get_state_manager()
    last_date = state.getdate("daily", name)
    state.setdate("daily", name, today)
    state.save()
    return last_date != today


Launch = Union[str, Callable]


@dataclass
class SmartStartEntry:
    """A smart start entry, which will only run when the condition is true

    :param str name: the name of the entry
    :param Union[Launch, List[Launch]] launch: the launch item
    :param Callable condition: the condition to check
    """

    name: str
    launch: Union[Launch, List[Launch]]
    condition: Callable[[], bool]

    def __call__(self):
        if self.condition and not self.condition():
            return
        self.run_anyway()

    def run_anyway(self):
        """Runs the launch item regardless of the condition"""

        def launch(launch):
            if isinstance(launch, str):
                if not is_exe_running(launch):
                    os.startfile(launch)
            elif callable(launch):
                launch()

        if isinstance(self.launch, list):
            for item in self.launch:
                launch(item)
        else:
            launch(self.launch)


class SmartStartManager:
    _smartstarts: Dict[str, SmartStartEntry]

    def __init__(self):
        self._smartstarts = dict()

    def register(
        self,
        entry=SmartStartEntry,
    ):
        """Registers a smart start entry"""
        if entry.name in self._smartstarts:
            raise ValueError(f"smart_start: {name} already exists")
        self._smartstarts[entry.name] = entry
        entry()

    def get_all(self):
        """Returns all smart start entries"""
        return self._smartstarts.values()


smartstart = SmartStartManager()
register_smartstart = smartstart.register
get_smartstarts = smartstart.get_all

if __name__ == "__main__":
    # open_chrome_fav_folder("bookmark_bar", "daily")
    print("daily once", daily_once("test"))
    print("daily once", daily_once("test"))



import sys

sys.coinit_flags = 0x0
import io
import os.path
import time
from ctypes.wintypes import DWORD, HWND, LONG
from datetime import datetime
from threading import Thread
from tkinter import messagebox
from traceback import print_exception
from typing import Callable, Dict, List, Optional, Sequence, Union

import pystray
from PIL import Image

from jigsawwm.hotkey import hotkey, keyboard_event_handler
from jigsawwm.manager import WindowManager
from jigsawwm.services import get_services
from jigsawwm.smartstart import get_smartstarts
from jigsawwm.w32.hook import Hook
from jigsawwm.w32.vk import Vk
from jigsawwm.w32.window import Window, is_app_window, is_window
from jigsawwm.w32.winevent import WinEvent


class Daemon:
    """JigsawWM Daemon serivce, you must inherite this class and override the `setup` function
    to configurate the Manager.
    """

    _hook: Hook
    _timers: Dict[Callable, Thread] = {}
    _timer_thread: Thread
    _trayicon: pystray.Icon
    _started: bool = False
    _wm: WindowManager
    menu_items: Sequence[pystray.MenuItem] = []

    def __init__(self):
        self._hook = Hook()
        self._timers = {}
        self._timer_thread = None
        self._trayicon = None

    def _error_handler(self, e: Exception):
        file = io.StringIO()
        # print_exception(e, e, None, file=file)
        print_exception(*sys.exc_info(), file=file)
        text = file.getvalue()
        print(text, file=sys.stderr)
        messagebox.showerror("JigsawWM", text)

    def hotkey(
        self,
        combkeys: Union[Sequence[Vk], str],
        target: Union[Callable, str],
        swallow: bool = True,
    ):
        """Register a global hotkey

        :param Union[Sequence[Vk], str] combkeys: key combination, i.e. ``[Vk.WIN, Vk.J]`` or ``Win + j``
        :param Union[Callback, str] target: can be a callback function or another key combination
        :param bool swallow: stop event propagation to prevent combkeys being processed by other programs
        """
        try:
            hotkey(combkeys, target, swallow, self._error_handler)
        except Exception as e:
            self._error_handler(e)

    def start_hooks(self):
        """Start all hooks"""
        self._hook.install_keyboard_hook(keyboard_event_handler)
        self._hook.install_winevent_hook(
            self._winevent_callback,
            WinEvent.EVENT_OBJECT_SHOW,
            WinEvent.EVENT_OBJECT_HIDE,
        )
        self._hook.install_winevent_hook(
            self._winevent_callback,
            WinEvent.EVENT_OBJECT_CLOAKED,
            WinEvent.EVENT_OBJECT_UNCLOAKED,
        )
        self._hook.install_winevent_hook(
            self._winevent_callback,
            WinEvent.EVENT_SYSTEM_MINIMIZESTART,
            WinEvent.EVENT_SYSTEM_MINIMIZEEND,
        )
        self._hook.install_winevent_hook(
            self._winevent_callback, WinEvent.EVENT_SYSTEM_MOVESIZEEND
        )
        self._hook.start()

    def _winevent_callback(
        self,
        event: WinEvent,
        hwnd: HWND,
        id_obj: LONG,
        id_chd: LONG,
        id_evt_thread: DWORD,
        evt_time: DWORD,
    ):
        if (
            id_obj
            or id_chd
            or not is_window(hwnd)
            or not is_app_window(hwnd)
            or self._wm.is_ignored(Window(hwnd))
        ):
            return
        print(
            "[{now}] {event:30s} {hwnd:8d} {title}".format(
                now=datetime.now().strftime("%M:%S.%f"),
                event=event.name,
                hwnd=hwnd,
                #  ido: {id_obj:6d} idc: {id_chd:6d}
                # id_obj=id_obj,
                # id_chd=id_chd,
                title=Window(hwnd).title,
            )
        )
        self._wm.sync(restrict=event == WinEvent.EVENT_SYSTEM_MOVESIZEEND)

    def stop_hooks(self):
        """Stop all hooks"""
        self._hook.stop()

    def timer(self, interval: float, callback: Callable, once: Optional[bool] = False):
        """Run callback function with a fixed time interval repeatedly

        :param float interval: interval between calls
        :param Callable callback: function to be called
        :param bool once: function would be called only once
        """

        # wrap func with in try-catch for safty
        def run():
            # global _timers
            while callback in self._timers:
                time.sleep(interval)
                try:
                    callback()
                except Exception as e:
                    self._error_handler(e)
                if once:
                    del self._timers[callback]

        self._timers[callback] = Thread(target=run)

    def start_timers(self):
        """Start all timers"""
        for timer_thread in self._timers.values():
            timer_thread.start()

    def stop_timers(self):
        """Stop all timers"""
        self._timers.clear()

    def start_trayicon(self):
        """Start trayicon"""
        script_dir = os.path.dirname(__file__)
        icon_path = os.path.join(script_dir, "assets", "logo.png")
        icon = Image.open(icon_path)

        def dynamic_menu() -> List[pystray.MenuItem]:
            service_menu_items = [
                pystray.MenuItem(
                    f"[{service.status_text}] {service.name}",
                    service.toggle,
                )
                for service in get_services()
            ]
            smartstart_menu_items = [
                pystray.MenuItem(
                    f"{smartstart.name}",
                    smartstart.run_anyway,
                )
                for smartstart in get_smartstarts()
            ]
            return [
                *self.menu_items,
                *smartstart_menu_items,
                pystray.Menu.SEPARATOR,
                *service_menu_items,
                pystray.Menu.SEPARATOR,
                pystray.MenuItem("Exit", self.stop),
            ]

        tray_icon = pystray.Icon(
            "JigsawWM",
            icon=icon,
            menu=pystray.Menu(dynamic_menu),
        )
        self._trayicon = tray_icon
        self._trayicon.run_detached()

    def stop_trayicon(self):
        """Stop trayicon"""
        self._trayicon.stop()
        self._trayicon = None

    def setup(self):
        """To be overrided by the users to configure the Window Manager"""
        pass

    def start(self):
        """Start daemon service"""
        try:
            self._wm = self.setup()
        except Exception as e:
            self._error_handler(e)
            sys.exit(1)
            return
        self._started = True
        self.start_trayicon()
        self.start_timers()
        self.start_hooks()
        while self._started:
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                self.stop()
                break

    def stop(self):
        """Stop daemon service"""
        self.stop_hooks()
        self.stop_timers()
        self.stop_trayicon()
        self._started = False



import enum
from concurrent.futures import ThreadPoolExecutor
from traceback import print_exception
from typing import Callable, Dict, FrozenSet, Iterator, Optional, Sequence, Tuple, Union

from jigsawwm.w32.hook import KBDLLHOOKDATA, KBDLLHOOKMSGID, Hook
from jigsawwm.w32.sendinput import (
    INPUT,
    INPUTTYPE,
    KEYBDINPUT,
    KEYEVENTF,
    is_synthesized,
    send_input,
)
from jigsawwm.w32.vk import Vk


def is_power_of_2(num: int) -> bool:
    return num != 0 and num & (num - 1) == 0


class Modifier(enum.IntFlag):
    """Keyboard modifier"""

    LCONTROL = enum.auto()
    LMENU = enum.auto()
    LSHIFT = enum.auto()
    LWIN = enum.auto()
    RCONTROL = enum.auto()
    RMENU = enum.auto()
    RSHIFT = enum.auto()
    RWIN = enum.auto()
    CONTROL = LCONTROL | RCONTROL
    MENU = LMENU | RMENU
    SHIFT = LSHIFT | RSHIFT
    WIN = LWIN | RWIN

    @classmethod
    def unfold(cls, mk: Union[str, int]) -> Iterator["Modifier"]:
        if not mk:
            return
        if isinstance(mk, str):
            mk = cls[mk].value
        if is_power_of_2(mk):
            yield cls(mk)
            return
        for v in cls.__members__.values():
            if v >= cls.CONTROL:
                return
            if v & mk:
                yield cls(v)


def expand_combination(
    combkeys: Sequence[Vk],
    index: Optional[int] = 0,
) -> Iterator[Sequence[Vk]]:
    """Expand `Ctrl+s` to `LCtrl+s` and `RCtrl+s`, so on and so forth"""
    key = combkeys[index]
    if key.name in Modifier.__members__:
        is_last = index + 1 == len(combkeys)
        for mk in Modifier.unfold(key.name):
            new_combkeys = combkeys[:index] + [Vk[mk.name]]
            if is_last:
                yield new_combkeys
            else:
                yield from expand_combination(
                    new_combkeys + combkeys[index + 1 :], index + 1
                )
    else:
        yield combkeys


# print("MENU+s", list(expand_combination([Vk.MENU, Vk.S])))
# print("LMENU+s", list(expand_combination([Vk.LMENU, Vk.S])))
# print("MENU+SHIFT+s", list(expand_combination([Vk.MENU, Vk.SHIFT, Vk.S])))
# print("F1", list(expand_combination([Vk.F1])))
# exit()

# { combination: (func, swallow, counteract) }
_hotkeys: Dict[
    FrozenSet[Vk], Tuple[Callable, bool, bool, Callable[[Exception], None]]
] = {}
_executor = ThreadPoolExecutor()


def hotkey(
    combkeys: Union[Sequence[Vk], str],
    target: Union[Callable, str],
    swallow: bool = True,
    error_handler: Callable[[Exception], None] = print_exception,
):
    """Register a system hotkey
    check `jigsawwm.w32.vk.Vk` for virtual key names
    check `_vk_aliases` for key aliases

    :param combkeys: Sequence[VirtualKey] | str, virtual keys combination
        example: [Vk.LCONTROL, Vk.LSHIFT, Vk.S] or "LControl+LShift+s"
    :param target: Callable | str, one of the following action would be carried
        out based on the type of the target:
        Callable:   the function would be executed
        str:        the str would be treated as a combination and send accordingly.
        i.e. "RWin+Space"
    :param swallow: stop combination being process by other apps
    """
    global _hotkeys
    # parse combkeys if it is string
    if isinstance(combkeys, str):
        combkeys = parse_hotkey(combkeys)
    # check if combination valid
    if not combkeys:
        raise Exception("empty combination")
    # turn target to function if it is string
    counteract = False
    if isinstance(target, str):
        target = combination_input(target)
        counteract = True
    count = len(list(filter(lambda vk: vk.name not in Modifier.__members__, combkeys)))
    if count != 1:
        raise Exception("require 1 and only 1 triggering key")
    for ck in expand_combination(combkeys):
        _hotkeys[frozenset(ck)] = (target, swallow, counteract, error_handler)


_vk_aliases: Dict[str, Vk] = {
    "LCTRL": Vk.LCONTROL,
    "LALT": Vk.LMENU,
    "RCTRL": Vk.RCONTROL,
    "RALT": Vk.RMENU,
    "CTRL": Vk.CONTROL,
    "MENU": Vk.MENU,
    "-": Vk.OEM_MINUS,
    "=": Vk.OEM_PLUS,
    ";": Vk.OEM_1,
    "/": Vk.OEM_2,
    "`": Vk.OEM_3,
    "[": Vk.OEM_4,
    "\\": Vk.OEM_5,
    "]": Vk.OEM_6,
    "'": Vk.OEM_7,
    ",": Vk.OEM_COMMA,
    ".": Vk.OEM_PERIOD,
}


def parse_hotkey(combkeys: str) -> Sequence[Vk]:
    """Converts combination in plain text ("Ctrl+s") to Sequence[Vk] ([Vk.CONTROL, Vk.S])"""
    parsed = []
    if not combkeys:
        return parsed
    global _vk_aliases
    for key_name in combkeys.split("+"):
        key = None
        key_name = key_name.strip().upper()
        # try alias
        key = _vk_aliases.get(key_name)
        # try name
        if key is None:
            if key_name not in Vk.__members__:
                raise Exception(f"invalid key: {key_name}")
            key = Vk[key_name]
        parsed.append(key)
    return parsed


def combination_input(target: str) -> Callable:
    """Converts key combination in text format to a Callable function"""
    target = parse_hotkey(target)

    def callback():
        for key in target:
            send_input(INPUT(type=INPUTTYPE.KEYBOARD, ki=KEYBDINPUT(wVk=key)))
        for key in reversed(target):
            send_input(
                INPUT(
                    type=INPUTTYPE.KEYBOARD,
                    ki=KEYBDINPUT(wVk=key, dwFlags=KEYEVENTF.KEYUP),
                )
            )

    return callback


_modifier = Modifier(0)


def keyboard_event_handler(msgid: KBDLLHOOKMSGID, msg: KBDLLHOOKDATA) -> bool:
    """Handles keyboard events and call callback if the combination
    had been registered
    """
    global _hotkeys, _executor
    # skip key we sent out
    if is_synthesized(msg):
        return False
    global _modifier, _hotkeys
    vkey = Vk(msg.vkCode)
    # print(msgid.name, vkey.name, msg.dwExtraInfo)
    if vkey.name in Modifier.__members__:
        # update modifier state if
        if msgid == KBDLLHOOKMSGID.WM_KEYDOWN:
            _modifier |= Modifier[vkey.name]
        elif msgid == KBDLLHOOKMSGID.WM_KEYUP:
            _modifier &= ~Modifier[vkey.name]
    elif msgid == KBDLLHOOKMSGID.WM_KEYDOWN:
        # see if combination registered
        combination = frozenset(
            (*map(lambda m: Vk[m.name], Modifier.unfold(_modifier)), vkey)
        )
        fs = _hotkeys.get(combination)
        if fs is not None:
            func, swallow, counteract, error_handler = fs
            if counteract:
                # send key up for combination to avoid confliction if the func
                # call send_input
                for key in combination:
                    send_input(
                        INPUT(
                            type=INPUTTYPE.KEYBOARD,
                            ki=KEYBDINPUT(wVk=key, dwFlags=KEYEVENTF.KEYUP),
                        )
                    )

            # wrap func with in try-catch for safty
            def wrapped_func():
                try:
                    func()
                except Exception as e:
                    error_handler(e)

            _executor.submit(wrapped_func)
            return swallow


if __name__ == "__main__":
    import time

    def delay_hello():
        time.sleep(1)
        print("hello world")

    hotkey([Vk.LWIN, Vk.B], delay_hello, True)

    hook = Hook()
    hook.install_keyboard_hook(keyboard_event_handler)
    hook.start()

    while True:
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            hook.stop()
            break
